import os
from rich import get_console
from rich.prompt import Confirm, Prompt
import yaml
from pydantic import BaseModel
from typing import List, Optional

available_scopes = [
    "arkitekt:rekuest:imitate",
    "arkitekt:herre",
    "arkitekt:fakts",
    "arkitekt:mikro:read",
    "arkitekt:mikro:imitate",
    "arkitekt:unlok",
]


class SoftwareManifest(BaseModel):
    identifier: str
    version: str
    description: Optional[str]
    scopes: List[str]


def ensure_gitignore(app_directory: str):
    """Ensure that the .gitignore file is present

    Args:
        app_directory (str): _description_
    """
    config_path = os.path.join(app_directory, ".arkitekt")
    gitignore_path = os.path.join(config_path, ".gitignore")
    gitignore = """credentials.yaml"""
    with open(gitignore_path, "w") as f:
        f.write(gitignore)


def load_manifest(app_directory: str) -> SoftwareManifest:
    """Load the manifest from the app directory

    Args:
        app_directory (str): _description_

    Returns:
        [type]: _description_
    """
    config_path = os.path.join(app_directory, ".arkitekt")
    fakts_path = os.path.join(config_path, "manifest.yaml")

    with open(fakts_path, "r") as f:
        return SoftwareManifest(**yaml.safe_load(f))


def initialize_manifest(
    app_directory: str,
    identifier: str = None,
    version: str = None,
    description: str = None,
    silent=False,
):
    """Initialize a new Arkitekt app



    Args:
        console (Console): _description_
        app_directory (str): _description_
        refresh (_type_, optional): _description_. Defaults to None.
        silent (bool, optional): _description_. Defaults to False.
    """

    # Create the project directory

    # Create the arkitekt.json file
    console = get_console()
    config_path = os.path.join(app_directory, ".arkitekt")
    if not os.path.exists(config_path):
        os.makedirs(config_path)

    fakts_path = os.path.join(config_path, "manifest.yaml")
    os.path.join(config_path, ".gitignore")

    initialize_fakts = True

    manifest = None

    if os.path.isfile(fakts_path):
        console.print(
            f"There seems to have been a previous app initialized at {app_directory}"
        )
        initialize_fakts = (
            Confirm.ask("Do you want to refresh the Apps configuration? ")
            if not silent
            else False
        )
        if not initialize_fakts:
            console.print(f"Nothing done. Bye :smiley: ")
            return

        with open(fakts_path, "r") as f:
            manifest = SoftwareManifest(**yaml.safe_load(f))

    console.print(f"Initializing a new Configuration")

    app_identifier = (
        identifier
        if identifier
        else Prompt.ask(
            "Give your app a unique identifier",
            default=manifest.identifier if manifest else None,
        )
    )
    app_version = (
        version
        if version
        else Prompt.ask(
            "Give your app a version", default=manifest.version if manifest else None
        )
    )
    app_description = (
        description
        if description
        else Prompt.ask(
            "Give your app a description",
            default=manifest.description if manifest else None,
        )
    )

    console.log("Which scopes do you want your app to have?")
    scopes = []
    for scope in available_scopes:
        if Confirm.ask(
            f"App should have {scope} scope?",
            default=scope in manifest.scopes if manifest else False,
        ):
            scopes.append(scope)

    if not scopes:
        console.print("App needs at least one scope")
        return

    manifest = {
        "identifier": app_identifier,
        "version": app_version,
        "description": app_description,
        "scopes": scopes,
    }

    with open(fakts_path, "w") as f:
        yaml.dump(manifest, f)

    ensure_gitignore(app_directory)



import os
from rich import get_console
from rich.prompt import Confirm, Prompt
from .initialize import load_manifest, ensure_gitignore
from .connect import load_credentials
import yaml
from fakts import Fakts
from fakts.grants.remote import DeviceCodeGrant
from fakts.discovery.base import FaktsEndpoint
from fakts.discovery.static import StaticDiscovery
from arkitekt.cli.prod.run import import_directory_and_start
import asyncio


def run(path, silent=False):
    """Run the app

    Args:
        app_directory (str): _description_
        silent (bool, optional): _description_. Defaults to False.
    """
    if path == ".":
        app_directory = os.getcwd()
    else:
        app_directory = os.path.join(os.getcwd(), path)

    console = get_console()
    config_path = os.path.join(app_directory, ".arkitekt")
    manifest_path = os.path.join(config_path, "manifest.yaml")
    credentials_path = os.path.join(config_path, "credentials.yaml")

    initialize_fakts = True

    if not os.path.isfile(manifest_path):
        console.print("No manifest found. Please initialize the app first.")
        return

    if not os.path.isfile(credentials_path):
        console.print("No credentials found. Please connect the app first.")
        return

    if not silent:
        console.print(f"Loading manifest")

    manifest = load_manifest(app_directory)
    credentials = load_credentials(app_directory)

    endpoint, token = next(iter(credentials.items()))
    print(endpoint, token)
    asyncio.run(import_directory_and_start(path, token=token, endpoint=endpoint))



from importlib import import_module
import logging
import sys
import os
from arkitekt.apps.fakts import ArkitektFakts
from fakts import Fakts
from rich.console import Console
from arkitekt import Arkitekt
import yaml
import json

logger = logging.getLogger(__name__)


class Dump:
    def __init__(
        self,
        path=None,
        entrypoint="run",
    ) -> None:

        self.console = Console()
        if path == ".":
            self.watch_path = os.getcwd()
            sys.path.insert(0, self.watch_path)
        else:
            self.watch_path = os.path.join(os.getcwd(), path)

        self.console.print(f"watch_path: {self.watch_path}")

        self.module_path = f"{path}.{entrypoint}" if path != "." else f"{entrypoint}"
        self.module = import_module(self.module_path)
        os.environ["ARKITEKT_AGENT_DEBUG"] = "False"

    async def run(self):

        app = Arkitekt(fakts=ArkitektFakts(fakts_path=f"{self.watch_path}/fakts.yaml"))

        if not os.path.exists(f"{self.watch_path}/.arkitekt"):
            os.mkdir(f"{self.watch_path}/.arkitekt")

        async with app:
            x = json.loads(app.fakts.grant.json())
            with open(f"{self.watch_path}/.arkitekt/config.yaml", "w") as file:
                documents = yaml.dump(x, file)

            x = app.rekuest.definition_registry.dump()
            with open(f"{self.watch_path}/.arkitekt/definitions.yaml", "w") as file:
                documents = yaml.dump(x, file)


async def import_directory_and_dump(path="watch", entrypoint="run"):
    host = Dump(path=path)
    await host.run()



from datetime import datetime
from importlib import reload, import_module
import asyncio
import asyncio
import sys
import time
from arkitekt import Arkitekt
from rich.console import Console
from arkitekt.apps.fakts import ArkitektFakts
from fakts import Fakts
from fakts.grants.remote.static import StaticGrant
from fakts.discovery.static import StaticDiscovery
from rich.console import Console
from arkitekt import Arkitekt

try:
    from watchdog.observers import Observer
    from watchdog.events import (
        FileModifiedEvent,
        FileSystemEventHandler,
    )
except ImportError:
    print("Please install watchdog to use this feature")
    sys.exit(1)

try:

    import janus
except ImportError:
    print("Please install janus to use this feature")
    sys.exit(1)

import os
import threading

from rekuest.definition.registry import (
    get_default_definition_registry,
)


class QueueHandler(FileSystemEventHandler):
    def __init__(self, *args, sync_q=None, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.queue = sync_q

    def on_any_event(self, event):
        self.queue.put(event)
        self.queue.join()


def watcher(path, queue, event: threading.Event):
    event_handler = QueueHandler(sync_q=queue)
    observer = Observer()
    observer.schedule(event_handler, path, recursive=True)
    observer.start()

    while not event.is_set():
        time.sleep(1)

    print("Cancelled because threading event is set")


async def buffered_queue(queue, timeout=4):
    buffer = [None]

    async def iterator():
        while True:
            nana = await queue.get()
            queue.task_done()
            if isinstance(nana, FileModifiedEvent):
                if nana.src_path.lower().endswith((".py")):
                    buffer.append(nana)

    loop = asyncio.get_event_loop()
    loop.create_task(iterator())
    while True:
        await asyncio.sleep(timeout)
        yield buffer[len(buffer) - 1]
        buffer = [None]


class Host:
    def __init__(
        self,
        path=None,
        entrypoint="run",
        token=None,
        endpoint=None,
    ) -> None:

        if path == ".":
            self.watch_path = os.getcwd()
            sys.path.insert(0, self.watch_path)
        else:
            self.watch_path = os.path.join(os.getcwd(), path)

        self.module_path = f"{path}.{entrypoint}" if path != "." else f"{entrypoint}"
        self.module = import_module(self.module_path)
        self.provide_task = None
        self.console = Console()
        self.token = token
        self.endpoint = endpoint
        os.environ["ARKITEKT_AGENT_DEBUG"] = "True"

    async def reprovide(self, endpoint=None, token=None):
        registry = get_default_definition_registry()
        registry.reset()
        reload(self.module)

        try:
            if registry.has_definitions():
                app = Arkitekt(
                    fakts=ArkitektFakts(
                        grant=StaticGrant(
                            discovery=StaticDiscovery(
                                base_url=self.endpoint
                                or os.getenv("FAKTS_ENDPOINT_URL")
                            ),
                            token=self.token or os.getenv("FAKTS_TOKEN"),
                        )
                    )
                )

                async with app:
                    self.console.print(
                        f"[bold green] --------------- Providing {datetime.now()} ------------- [/bold green]"
                    )
                    await app.rekuest.run()

        except Exception as e:
            self.console.print(
                f"[bold red] --------------- Error in App {datetime.now()} ------------- [/bold red]"
            )
            self.console.print_exception()

        except asyncio.CancelledError as e:
            self.console.print(
                f"[bold yellow] --------------- App Cancelled {datetime.now()} ------------- [/bold yellow]"
            )
            raise e

    async def restart(self):
        loop = asyncio.get_event_loop()

        if self.provide_task:
            self.provide_task.cancel()
            try:
                await self.provide_task
            except asyncio.CancelledError as e:
                pass

        self.provide_task = loop.create_task(self.reprovide())

    async def run(self):
        await self.restart()

        jqueue = janus.Queue()

        cancel_event = threading.Event()
        loop = asyncio.get_running_loop()
        fut = loop.run_in_executor(
            None, watcher, self.watch_path, jqueue.sync_q, cancel_event
        )

        try:
            async for event in buffered_queue(jqueue.async_q, timeout=1):
                if event:
                    self.console.print(
                        f"[bold yellow] --------------- File Changed {datetime.now()} ------------- [/bold yellow]"
                    )
                    await self.restart()
        except asyncio.CancelledError as e:
            cancel_event.set()
            await fut

        jqueue.close()
        await jqueue.wait_closed()


async def watch_directory_and_stage(path="watch", token=None, endpoint=None):
    host = Host(path=path, token=token, endpoint=endpoint)
    await host.run()
